######################################################################
#
# File: test/unit/_cli/fixtures/module_loading_b2sdk.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

# This is a helper module for test_autocomplete_cache.py

from b2sdk.v2 import B2Api  # noqa


def function():
    pass
