######################################################################
#
# File: test/integration/__init__.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
"""
B2 CLI integrations tests

This package contains tests that require interaction with remote server.
Integration tests should be runnable against any arbitrary command implementing `b2` CLI, supplied via `--sut` flag.
"""
